package fmon.builder

import java.io._

import javafx.application.Application
import javafx.fxml.FXML
import javafx.fxml.FXMLLoader
import javafx.fxml.JavaFXBuilderFactory
import javafx.scene.{control => jfxsc, Parent, Scene}
import javafx.stage.Stage

import scalafx.Includes._
import scalafx.beans.property._
import scalafx.collections.ObservableBuffer
import scalafx.scene.control._
import scalafx.scene.image.ImageView
import scalafx.scene.input.MouseEvent
import scalafx.scene.paint.Color
import scalafx.scene.layout._
import scalafx.util.Duration
import scalafx.util.StringConverter

import fmon.draw._
import fmon.util.YamlHelper

class ObsAnimation(name_ : String) {
  val name = new StringProperty(this, "name", name_)
  val frames = new ObservableBuffer[Frame]()
  
  def toToken = {
    val filename = frames(0).file.getName
    val indices = frames.map(_.index).toIndexedSeq
    AnimationToken(filename, indices)
  }
  
  override def toString = name()
}

class Frame(val img: StillImg, val file: File, val index: Int) {
  def toToken = ImageToken(file.getName, index)
  
  override def toString = s"${file.getName} ${index}"
}



case class ImageToken(val filename: String, val index: Int) {
  def toFrame(dir: File) = {
    val file = new File(s"${dir.getPath}/$filename")
    val pal = Palette.square(new FileInputStream(file), 5).images(5)
  }
  override def toString = s"${filename} ${index}"
}

class AnimationBuilder extends Savable {
  @FXML var animationList: jfxsc.ListView[ObsAnimation] = _
  @FXML var frameList: jfxsc.ListView[Frame] = _
  @FXML var viewpane: javafx.scene.image.ImageView = _
  @FXML var tiles: javafx.scene.layout.TilePane = _
  @FXML var fileChooser: jfxsc.ComboBox[File] = _
  
  val animations = ObservableBuffer[ObsAnimation]()
  val imgFiles = ObservableBuffer[File]()
  
  val imgdir = raw"C:\Users\dalyj\Documents\Design\Images\TimmahLexusX reduced\Reduced Animations"
  
  @FXML
  def initialize(): Unit = {
    fileChooser.items = imgFiles
    fileChooser.selectionModel().selectedIndex.onChange(selectImageFile)
    
    fileChooser.converter = FilenameConverter
    val dir = new File(imgdir)
    imgFiles ++= dir.listFiles()
    
    animationList.items = animations
    animationList.selectionModel().selectedIndex.onChange(selectionChange)
    frameList.selectionModel().selectedIndex.onChange(selectFrame)
    //for (x <- 0 until palette.numAcross; y <- 0 until palette.numDown) {
    
    //addAnimation()
    animationList.selectionModel().selectFirst()
    fileChooser.selectionModel().selectFirst()
  }
  
  def currFile = fileChooser.selectionModel().getSelectedItem
  def palette = Palette.square(new FileInputStream(currFile), 5)
  
  def selectImageFile(): Unit = {
    val images = palette.images
    val children = images.zipWithIndex.map{case (img, i) => { 
      val imgView = new ImageView() {
        scaleX = 0.5
        scaleY = 0.5
        image = img.croppedImage()
        
      }
      imgView.handleEvent(MouseEvent.Any) {
        e: MouseEvent => if (e.clickCount == 2 && e.eventType == MouseEvent.MouseClicked) {
          addFrame(img, i)
        }
      }
      imgView.delegate
    }}
    tiles.children.clear()
    tiles.children ++= children
    tiles.prefTileWidth = 96 // TODO : I think this has to do with how the image is scaled
    tiles.prefTileHeight = 96
  }
  
  def addAnimation(): Unit = {
    addAnimation("Animation")
  }
  
  def addAnimation(name: String): Unit = {
    animations += new ObsAnimation(name)
    animationList.selectionModel().selectLast()
  }
  
  def quickBuildAnimation(): Unit = {
    val filename = currFile.getName
    val name = filename.split('.')(0)
    animations += new ObsAnimation(name)
    animationList.selectionModel().selectLast()
    palette.images.zipWithIndex.foreach{case (img, i) => addFrame(img, i)}
  }
  
  def selectionChange(): Unit = {
    frameList.items = selected.frames
  }
  
  def playAnimation(): Unit = {
    val images = frameList.items().map(f => f.img).toIndexedSeq
    val animated = new AnimatedImage(images)
    val duration = new Duration(50 * images.size)
    val animator = new SpriteAnimation(viewpane, animated, duration)
    animator.play()
  }
  
  def addFrame(): Unit = {
    //println("New frame")
    //selected.frames += new FrameToken
  }
  
  def addFrame(img: StillImg, index: Int): Unit = {
    selected.frames += new Frame(img, currFile, index)
  }
  
  def removeFrame(): Unit = {
    frameList.items() --= frameList.selectionModel().getSelectedItems
  }
  
  def selectFrame(): Unit = {
    val frame = frameList.selectionModel().getSelectedItem
    frame.img.draw(viewpane, 0)
  }
  
  override def saveTo(file: File): Unit = {
    val aniMap = animations.map(a => (a.name(), a.toToken)).toMap
    YamlHelper.writeMap(new FileOutputStream(file), aniMap)
  }
  
  override def openFrom(file: File): Unit = {
    val aniMap = YamlHelper.extractMap[AnimationToken](new FileInputStream(file))
    aniMap.foreach{case (name, token) => {
      addAnimation(name)
      val frames = AnimationBuilder.toFrames(token, new File(imgdir))
      animationList.selectionModel().getSelectedItem.frames ++= frames
    }}
  }
  
  private def selected: ObsAnimation = {
    animationList.selectionModel().getSelectedItem
  }
}

object AnimationBuilder {
  class ABA extends Application {
    override def start(primaryStage: Stage): Unit = {
      val frameLoader = new FXMLLoader(getClass.getResource("App.fxml"))
      val root: Parent = frameLoader.load()
      val controller = frameLoader.getController[App]()
      
      val builderLoader = new FXMLLoader(getClass.getResource("AnimationBuilder.fxml"))
      val builder: Parent = builderLoader.load()
      val builderController = builderLoader.getController[AnimationBuilder]()
      
      controller.pane.children = builder
      controller.builder = builderController
      
      val scene: Scene = new Scene(root)
      primaryStage.setScene(scene)
      primaryStage.show()
    }
  }
  
  def main(args: Array[String]): Unit = {
    Application.launch(classOf[ABA], args: _*)
  }
  
  def toFrames(token: AnimationToken, dir: File) = {
    val file = new File(s"${dir.getPath}/${token.filename}")
    val palette = Palette.square(new FileInputStream(file), 5)
    val images = palette.images
    token.indices.map(i => new Frame(images(i), file, i))
  }
}