package fmon.builder

import java.io._

import javafx.application.Application
import javafx.fxml.FXML
import javafx.fxml.FXMLLoader
import javafx.fxml.JavaFXBuilderFactory
import javafx.scene.{control => jfxsc, Parent, Scene}
import javafx.stage.Stage

import scalafx.Includes._
import scalafx.beans.property._
import scalafx.collections.ObservableBuffer
import scalafx.scene.control._
import scalafx.scene.paint.Color
import scalafx.scene.layout._

import org.yaml.snakeyaml.nodes._
import org.yaml.snakeyaml.nodes.Tag
import org.yaml.snakeyaml.constructor._
import org.yaml.snakeyaml.representer._

import fmon.stat.{Effectiveness, Element}
import fmon.util.YamlHelper
import Effectiveness._

import TableColumn._

class ObsElement(name_ : String, bgColor_ : Color = Color.Gray, fontColor_ : Color = Color.Black) {
  val name = new StringProperty(this, "name", name_)
  val bgColor = new ObjectProperty[Color](this, "bgColor", bgColor_)
  val fontColor = new ObjectProperty[Color](this, "fontColor", fontColor_)
  
  val immune = new ObservableBuffer[ObsElement]()
  val resist = new ObservableBuffer[ObsElement]()
  val normal = new ObservableBuffer[ObsElement]()
  val weak = new ObservableBuffer[ObsElement]()
  
  def asElement = {
    def toMap(buffer: ObservableBuffer[ObsElement], mult: Effectiveness) = {
      buffer.map(e => (e.name(), mult)).toMap
    }
    val effect = toMap(immune, Immune) ++ toMap(resist, Resist) ++ toMap(normal, Regular) ++ toMap(weak, Weak)  
    new Element(name(), bgColor(), fontColor(), effect)
  }
  
  override def toString = name()
}

class ColorRepresenter extends Representer {
  representers.put(classOf[Color], new ColorRepresent)
  
  class ColorRepresent extends Represent {
    override def representData(data: AnyRef): Node = {
      val color = data.asInstanceOf[Color]
      representScalar(new Tag("!color"), color.toString())
    }
  }
}

class ColorConstructor extends Constructor {
  yamlConstructors.put(new Tag("!color"), new ConstructColor)
  
  class ConstructColor extends AbstractConstruct {
    override def construct(node: Node): AnyRef = {
      val string = constructScalar(node.asInstanceOf[ScalarNode])
      return Color.valueOf(string)
    }
  }
}

class ElementBuilder extends Savable {
  @FXML var elements: jfxsc.ListView[ObsElement] = _
  @FXML var add: jfxsc.Button = _
  //@FXML var matchups: jfxsc.TableView[ObsElement] = _
  var elementList = ObservableBuffer[ObsElement]()
  
  @FXML var immuneView: jfxsc.ListView[ObsElement] = _
  @FXML var resistView: jfxsc.ListView[ObsElement] = _
  @FXML var normalView: jfxsc.ListView[ObsElement] = _
  @FXML var weakView: jfxsc.ListView[ObsElement] = _
  
  @FXML var name: jfxsc.TextField = _
  @FXML var bgColor: jfxsc.ColorPicker = _
  @FXML var textColor: jfxsc.ColorPicker = _
  @FXML var button: jfxsc.Button = _
  
  @FXML
  def initialize(): Unit = {
    elements.items = elementList 
    elementList ++= Seq(new ObsElement("Fire"), new ObsElement("Water"), new ObsElement("Leaf"))
    elements.selectionModel().selectedIndex.onChange(selectionChange)
    elements.selectionModel().selectFirst()
    
    elementList.foreach(e => elementList.foreach(f => e.normal += f))
  }
  
  def selectionChange(): Unit = {
    val element = elementList(elements.selectionModel().selectedIndex())
    name.text = element.name()
    bgColor.value = element.bgColor()
    textColor.value = element.fontColor()
    
    immuneView.items = element.immune
    resistView.items = element.resist
    normalView.items = element.normal
    weakView.items = element.weak
    updateFont()
  }
  
  def updateName(): Unit = {
    button.text = name.text()
    setChanges()
  }
  
  def updateBg(): Unit = {
    updateFont()
  }
  
  def updateFont(): Unit = {
    val bg = bgColor.value().toString().replaceFirst("0x", "#")
    val color = textColor.value().toString().replaceFirst("0x", "#")
    button.style = s"-fx-background-color: ${bg}; -fx-text-fill: ${color}"
    setChanges()
  }
  
  def setChanges(): Unit = {
    val element = elementList(elements.selectionModel().selectedIndex())
    element.name.value = name.text()
    element.bgColor.value = bgColor.value()
    element.fontColor.value = textColor.value()
    elements.refresh()
  }
  
  def addElement(): Unit = {
    elementList += new ObsElement("New Element")
    println(elements.items().size())
  }
  
  def immune2Resist(): Unit = {
    val items = immuneView.selectionModel().getSelectedItems
    resistView.items() ++= items
    immuneView.items() --= items
  }
  
  def resist2Immune(): Unit = {
    val items = resistView.selectionModel().getSelectedItems
    immuneView.items() ++= items
    resistView.items() --= items
  }
  
  def resist2Normal(): Unit = {
    val items = resistView.selectionModel().getSelectedItems
    normalView.items() ++= items
    resistView.items() --= items
  }
  
  def normal2Resist(): Unit = {
    val items = normalView.selectionModel().getSelectedItems
    resistView.items() ++= items
    normalView.items() --= items
  }
  
  def normal2Weak(): Unit = {
    val items = normalView.selectionModel().getSelectedItems
    weakView.items() ++= items
    normalView.items() --= items
  }
  
  def weak2Normal(): Unit = {
    val items = weakView.selectionModel().getSelectedItems
    normalView.items() ++= items
    weakView.items() --= items
  }
  
  override def saveTo(file: File): Unit = {
    val elementMap = elementList.map(e => (e.name(), e.asElement)).toMap
    YamlHelper.writeMap(new FileOutputStream(file), elementMap)
  }
  
  override def openFrom(file: File): Unit = {
    val elementMap = YamlHelper.extractMap[Element](new FileInputStream(file))
    val es = elementMap.values.map(e => new ObsElement(e.name, e.bgColor, e.fontColor)).toSeq
    es.foreach(eo => es.foreach(fo => {
      val e = elementMap(eo.name())
      val f = elementMap(fo.name())
      e.effect.getOrElse(f.name, Regular) match {
        case Immune => eo.immune += fo
        case Resist => eo.resist += fo
        case Regular => eo.normal += fo
        case Weak => eo.weak += fo
        case _ => throw new IllegalArgumentException
      }
    }))
    elementList.clear()
    elementList ++= es
    elements.selectionModel().selectFirst()
  }
}

object ElementBuilder {
  class EBA extends Application {
    override def start(primaryStage: Stage): Unit = {
      val frameLoader = new FXMLLoader(getClass.getResource("App.fxml"))
      val root: Parent = frameLoader.load()
      val controller = frameLoader.getController[App]()
      
      val builderLoader = new FXMLLoader(getClass.getResource("ElementBuilder.fxml"))
      val builder: Parent = builderLoader.load()
      val builderController = builderLoader.getController[ElementBuilder]()
      
      controller.pane.children = builder
      controller.builder = builderController
      
      val scene: Scene = new Scene(root)
      primaryStage.setScene(scene)
      primaryStage.show()
    }
  }
  
  def main(args: Array[String]): Unit = {
    Application.launch(classOf[EBA], args: _*)
  }
}