package fmon.builder

import java.io._
import java.util.prefs.Preferences

import scala.util.Properties

import javafx.application.Application
import javafx.fxml.FXML
import javafx.fxml.FXMLLoader
import javafx.fxml.JavaFXBuilderFactory
import javafx.scene.{control => jfxsc, layout => jfxsl, Parent, Scene}
import javafx.stage.Stage

import scalafx.Includes._
import scalafx.beans.property._
import scalafx.collections.ObservableBuffer
import scalafx.scene.control._
import scalafx.scene.image.ImageView
import scalafx.scene.input.MouseEvent
import scalafx.scene.paint.Color
import scalafx.scene.layout._
import scalafx.stage.FileChooser
import scalafx.stage.Modality
import FileChooser.ExtensionFilter

import fmon._
import fmon.builder.dialog.MapProperties
import fmon.draw.Palette
import fmon.draw.tile._
import fmon.util.{Direction, YamlHelper}
import fmon.world._

import Direction._
import Alert.AlertType

import scala.math.{min, max}

sealed trait MapBuilderState

case object GroundTile extends MapBuilderState
case object DoodadTile extends MapBuilderState

class MapBuilder {
  @FXML var tileSelector: jfxsl.TilePane = _
  @FXML var doodadSelector: jfxsl.TilePane = _
  @FXML var gameMap: jfxsl.TilePane = _
  @FXML var doodadMap: jfxsl.TilePane = _
  
  var imgSeq: IndexedSeq[ImageView] = _
  var doodadSeq: IndexedSeq[ImageView] = _
  
  val tilesets = {
    val dir = new File(Config.homedir + raw"Progena\Data")
    YamlHelper.extractSeq[TilesetToken](new FileInputStream(new File(dir, "Tilesets.yaml"))).map(t => (t.name, t)).toMap
  }
  var tileset: Tileset = _
  var currTile: AutoTile = _
  var currState: MapBuilderState = GroundTile
  
  var level: GameMap = _
  
  var lastFile: File = _
  
  def lastDir(): String = {
    val prefs = Preferences.userNodeForPackage(classOf[MapBuilder])
    prefs.get("dir", Properties.userDir)
  }
  
  def rememberDir(file: File): Unit = {
    val dir = if (file.isDirectory()) file.getPath else file.getParent
    if (dir != lastDir) {
      val prefs = Preferences.userNodeForPackage(classOf[MapBuilder])
      prefs.put("dir", dir)
    }
  }
  
  @FXML
  def initialize(): Unit = {
    val resourceDir = new File(Config.homedir + raw"Progena\Resources\tilesets")
    tileset = tilesets("Outside").load(resourceDir)
    level = new GameMap(10, 10, tileset)
    setup()
  }
  
  def setup(): Unit = {
    
    val icons = tileset.groundTiles.map(t => {
      val view = new ImageView(t.icon.croppedImage())
      view.handleEvent(MouseEvent.Any) {
        e: MouseEvent => if (e.eventType == MouseEvent.MouseClicked) {
          currTile = t
          currState = GroundTile
        }
      }
      view.delegate
    })
    tileSelector.children.clear()
    tileSelector.children ++= icons
    currTile = tileset.groundTiles.head
    
    val doodads = tileset.doodadTiles.map(d => {
      val view = new ImageView(d.icon.croppedImage())
      view.handleEvent(MouseEvent.Any) {
        e: MouseEvent => if (e.eventType == MouseEvent.MouseClicked) {
          currTile = d
          currState = DoodadTile
        }
      }
      view.delegate
    })
    doodadSelector.children.clear()
    doodadSelector.children ++= doodads 
    
    imgSeq = for (y <- 0 until level.height; x <- 0 until level.width) yield {
      val view = new ImageView(level(x, y).icon.croppedImage()) // TODO Tile with adjacent
      view.handleEvent(MouseEvent.Any) {
        e: MouseEvent => if (e.eventType == MouseEvent.MouseClicked) {
          val index = point2index(x, y)
          currState match {
            case GroundTile => {
              level.tiles = level.tiles.updated(index, currTile)
              autotile(x, y)
            }
            case DoodadTile => {
              level.doodads = level.doodads.updated(index, currTile)
              doodadSeq(index).image = level.doodads(index).icon.croppedImage()
            }
          }
          
        }
      }
      view
    }
    doodadSeq = for (y <- 0 until level.height; x <- 0 until level.width) yield {
      val view = new ImageView(level.doodad(x, y).icon.croppedImage())
      view
    }
    gameMap.children.clear()
    gameMap.prefColumns = level.width
    gameMap.prefRows = level.height
    gameMap.children ++= imgSeq.map(_.delegate)
    doodadMap.children.clear()
    doodadMap.prefColumns = level.width
    doodadMap.prefRows = level.height
    doodadMap.children ++= doodadSeq.map(_.delegate)
    doodadMap.mouseTransparent = true
    
    for (y <- 0 until level.height; x <- 0 until level.width) {
      smoothTile(x, y)
    }
  }
  
  def autotile(x: Int, y: Int) = {
    for (yy <- max(y - 1, 0) to min(y + 1, level.height - 1); 
        xx <- max(x - 1, 0) to min(x + 1, level.width - 1)) {
      smoothTile(xx, yy)
    }
  }
  
  def smoothTile(x: Int, y: Int) = {
    imgSeq(level.pt2index(x, y)).image = level.smoothed(x, y)
  }
  
  @FXML
  def onPropertiesDialog: Unit = {
    val fxmlLoader = new FXMLLoader(getClass().getResource("dialog/MapProperties.fxml"))
    val parent: Parent = fxmlLoader.load()
    val dialogController = fxmlLoader.getController[MapProperties]();
        //dialogController.setAppMainObservableList(tvObservableList);
    dialogController.init(level)

    val dialog = new Alert(AlertType.Confirmation) {
      title = "Map Properties"
      dialogPane().content = parent
      headerText = ""
      graphic = null
    }
    val result = dialog.showAndWait()
    println(result)
    result match {
      case Some(ButtonType.OK) => {
        println(dialogController.width, dialogController.height)
        level = new GameMap(dialogController.width, dialogController.height, tileset)
        level.tiles = level.tiles.map(_ => currTile)
        setup()
      }
      case _ => ()
    }
    /*val scene = new Scene(parent, 300, 200);
    val stage = new Stage();
    stage.initModality(Modality.ApplicationModal);
    stage.setScene(scene);
    stage.showAndWait();*/
  }
  
  def save() = {
    if (lastFile != null) {
      level saveTo lastFile
    } else {
      saveAs()
    }  
  }
  
  def saveAs() = {
    val fileChooser = new FileChooser {
      title = "Open Resource File"
      initialDirectory = new File(lastDir())
      extensionFilters ++= Seq(
        new ExtensionFilter("MAP Files", "*.map"),
        new ExtensionFilter("All Files", "*.*")
      )
    }
    
    val selectedFile = fileChooser.showSaveDialog(null)
    if (selectedFile != null) {
      rememberDir(selectedFile)
      level saveTo selectedFile
      lastFile = selectedFile
    }
  }
  
  def open() = {
    val fileChooser = new FileChooser {
      title = "Open Resource File"
      initialDirectory = new File(lastDir())
      extensionFilters ++= Seq(
        new ExtensionFilter("MAP Files", "*.map"),
        new ExtensionFilter("All Files", "*.*")
      )
    }
    
    val selectedFile = fileChooser.showOpenDialog(null)
    if (selectedFile != null) {
      rememberDir(selectedFile)
      level = GameMap loadFrom selectedFile
      tileset = level.tileset
      setup()
      lastFile = selectedFile
    }
  }
  
  def point2index(x: Int, y: Int) = level.pt2index(x, y)
  
  def left(i: Int) = i - 1
  def right(i: Int) = i + 1
  def up(i: Int) = i - level.width
  def down(i: Int) = i + level.width
}

object MapBuilder {
  class MBA extends Application {
    override def start(primaryStage: Stage): Unit = {
      val frameLoader = new FXMLLoader(getClass.getResource("MapBuilder.fxml"))
      val root: Parent = frameLoader.load()
      val controller = frameLoader.getController[MapBuilder]()
      
      //val builderLoader = new FXMLLoader(getClass.getResource("ElementBuilder.fxml"))
      //val builder: Parent = builderLoader.load()
      //val builderController = builderLoader.getController[ElementBuilder]()
      
      //controller.pane.children = builder
      //controller.builder = builderController
      
      val scene: Scene = new Scene(root)
      primaryStage.setScene(scene)
      primaryStage.width = 1280
      primaryStage.height = 800
      primaryStage.show()
    }
  }
  
  def main(args: Array[String]): Unit = {
    Application.launch(classOf[MBA], args: _*)
  }
}