package fmon.draw.tile

import java.io.{File, FileInputStream}

import scalafx.scene.SnapshotParameters
import scalafx.scene.canvas.Canvas
import scalafx.scene.image._
import scalafx.scene.paint.Color

import fmon._
import fmon.draw._
import fmon.util._

import Direction._

trait AutoTile {
  def icon: StillImg
  def build(dirs: Set[Direction]): StillImg
}

class BasicTile(val icon: StillImg) extends AutoTile {
  def build(dirs: Set[Direction]) = icon
}

class AutoFloorTile(val palette: Palette, val size: Int = 48) extends AutoTile {
  
  def icon: StillImg = palette.apply(0, 0, 2, 2)
  
  def build(dirs: Set[Direction]) = {
    val canvas = new scalafx.scene.canvas.Canvas(size, size)
    val g = canvas.graphicsContext2D
    
    val ne = getNE(dirs)
    val nw = getNW(dirs)
    val se = getSE(dirs)
    val sw = getSW(dirs)
    
    canvas.graphicsContext2D.drawImage(nw, 0, 0)
    canvas.graphicsContext2D.drawImage(ne, size / 2, 0)
    canvas.graphicsContext2D.drawImage(sw, 0, size / 2)
    canvas.graphicsContext2D.drawImage(se, size / 2, size / 2)
    
    val wi = new WritableImage(size, size)
    val parameters = new SnapshotParameters(){
      fill = Color.Transparent
    }
    val img = canvas.snapshot(parameters, wi)
    StillImg(img, 0, 0, size, size)
  }
  
  private def getNE(dirs: Set[Direction]): Image = {
    if (!dirs(Northeast) && dirs(North) && dirs(East)) {
      palette(3, 0).croppedImage()
    } else {
      val x = if (dirs(East)) 1 else 3
      val y = if (dirs(North)) 4 else 2
      palette(x, y).croppedImage
    }
  }
  
  private def getNW(dirs: Set[Direction]): Image = {
    if (!dirs(Northwest) && dirs(North) && dirs(West)) {
      palette(2, 0).croppedImage()
    } else {
      val x = if (dirs(West)) 2 else 0
      val y = if (dirs(North)) 4 else 2
      palette(x, y).croppedImage
    }
  }
  
  private def getSE(dirs: Set[Direction]): Image = {
    if (!dirs(Southeast) && dirs(South) && dirs(East)) {
      palette(3, 1).croppedImage()
    } else {
      val x = if (dirs(East)) 1 else 3
      val y = if (dirs(South)) 3 else 5
      palette(x, y).croppedImage
    }
  }
  
  private def getSW(dirs: Set[Direction]): Image = {
    if (!dirs(Southwest) && dirs(South) && dirs(West)) {
      palette(2, 1).croppedImage()
    } else {
      val x = if (dirs(West)) 2 else 0
      val y = if (dirs(South)) 3 else 5
      palette(x, y).croppedImage
    }
  }
}

class AutoWallTile(val palette: Palette, val size: Int = 48) extends AutoTile {
  def icon: StillImg = {
    val canvas = new scalafx.scene.canvas.Canvas(size, size)
    val g = canvas.graphicsContext2D
    
    val ul = palette.apply(0, 0)
    val ur = palette.apply(3, 0)
    val ll = palette.apply(0, 3)
    val lr = palette.apply(3, 3)
    
    canvas.graphicsContext2D.drawImage(ul.croppedImage(), 0, 0)
    canvas.graphicsContext2D.drawImage(ur.croppedImage(), size / 2, 0)
    canvas.graphicsContext2D.drawImage(ll.croppedImage(), 0, size / 2)
    canvas.graphicsContext2D.drawImage(lr.croppedImage(), size / 2, size / 2)
    
    val wi = new WritableImage(size, size)
    val parameters = new SnapshotParameters(){
      fill = Color.Transparent
    }
    val img = canvas.snapshot(parameters, wi)
    StillImg(img, 0, 0, size, size)
  }
  
  def build(dirs: Set[Direction]) = {
    val canvas = new scalafx.scene.canvas.Canvas(size, size)
    val g = canvas.graphicsContext2D
    
    val ul = buildUL(dirs)
    val ur = buildUR(dirs)
    val ll = buildLL(dirs)
    val lr = buildLR(dirs)
    
    canvas.graphicsContext2D.drawImage(ul.croppedImage(), 0, 0)
    canvas.graphicsContext2D.drawImage(ur.croppedImage(), size / 2, 0)
    canvas.graphicsContext2D.drawImage(ll.croppedImage(), 0, size / 2)
    canvas.graphicsContext2D.drawImage(lr.croppedImage(), size / 2, size / 2)
    
    val wi = new WritableImage(size, size)
    val parameters = new SnapshotParameters(){
      fill = Color.Transparent
    }
    val img = canvas.snapshot(parameters, wi)
    StillImg(img, 0, 0, size, size)
  }
  
  private def buildUL(dirs: Set[Direction]) = {
    val x = if (dirs(West)) 2 else 0
    val y = if (dirs(North)) 2 else 0
    palette(x, y)
  }
  
  private def buildUR(dirs: Set[Direction]) = {
    val x = if (dirs(East)) 1 else 3
    val y = if (dirs(North)) 2 else 0
    palette(x, y)
  }
  
  private def buildLL(dirs: Set[Direction]) = {
    val x = if (dirs(West)) 2 else 0
    val y = if (dirs(South)) 1 else 3
    palette(x, y)
  }
  
  private def buildLR(dirs: Set[Direction]) = {
    val x = if (dirs(East)) 1 else 3
    val y = if (dirs(South)) 1 else 3
    palette(x, y)
  }
}

trait AutoTilePalette {
  def apply(x: Int, y: Int): AutoTile
  def tiles: IndexedSeq[AutoTile]
}

class AutoFloorTilePalette(val palette: Palette, val size: Int) extends AutoTilePalette {
  final val TileWidth = 2
  final val TileHeight = 3
  
  def this(file: File, size: Int = 48) = this(Palette.bySize(new FileInputStream(file), size, size), size)
  
  def apply(x: Int, y: Int) = {
    val img = palette.apply(x * TileWidth, y * TileHeight, TileWidth, TileHeight).croppedImage()
    new AutoFloorTile(new Palette(img, 2 * TileWidth, 2 * TileHeight), size)
  }
  
  def tiles = for (y <- 0 until palette.numDown / TileHeight; x <- 0 until palette.numAcross / TileWidth) yield {
    this(x, y)
  }
}

class AutoWallTilePalette(val palette: Palette, val size: Int) extends AutoTilePalette {
  final val TileWidth = 2
  final val TileHeight = 2
  
  def this(file: File, size: Int = 48) = this(Palette.bySize(new FileInputStream(file), size, size), size)
  
  def apply(x: Int, y: Int) = {
    val img = palette.apply(x * TileWidth, y * TileHeight, TileWidth, TileHeight).croppedImage()
    new AutoWallTile(new Palette(img, 2 * TileWidth, 2 * TileHeight), size)
  }
  
  def tiles = for (y <- 0 until palette.numDown / TileHeight; x <- 0 until palette.numAcross / TileWidth) yield {
    this(x, y)
  }
}

class AutoComboTilePalette(val palette: Palette, val size: Int) extends AutoTilePalette {
  final val TileWidth = 2
  final val CeilHeight = 3
  final val WallHeight = 2
  final val TotalHeight = CeilHeight + WallHeight
  
  def this(file: File, size: Int = 48) = this(Palette.bySize(new FileInputStream(file), size, size), size)
  
  def apply(x: Int, y: Int) = {
    val yy = y / 2
    if (y % 2 == 0) {
      // Ceiling
      val img = palette.apply(x * TileWidth, yy * TotalHeight, TileWidth, CeilHeight).croppedImage()
      new AutoFloorTile(new Palette(img, 2 * TileWidth, 2 * CeilHeight), size)
    } else {
      // Wall
      val img = palette.apply(x * TileWidth, yy * TotalHeight + CeilHeight, TileWidth, WallHeight).croppedImage()
      new AutoWallTile(new Palette(img, 2 * TileWidth, 2 * WallHeight), size)
    }
  }
  
  def tiles = for (y <- 0 until 2 * palette.numDown / TotalHeight; x <- 0 until palette.numAcross / TileWidth) yield {
    this(x, y)
  }
}

class BasicTilePalette(val palette: Palette, val size: Int) extends AutoTilePalette {
  def this(file: File, size: Int = 48) = this(Palette.bySize(new FileInputStream(file), size, size), size)
  
  def apply(x: Int, y: Int) = new BasicTile(palette(x, y))
  def tiles = palette.images.map(new BasicTile(_))
}

class PartitionedTilePalette(val palette: Palette, val parts: Int, val size: Int) extends AutoTilePalette {
  val subwidth = palette.numAcross / parts
  
  def this(file: File, parts: Int, size: Int = 48) = this(Palette.bySize(new FileInputStream(file), size, size), parts, size)
  
  def apply(x: Int, y: Int) = {
    val yy = y % palette.numDown
    val xx = (y / palette.numDown) * subwidth
    new BasicTile(palette(xx, yy))
  }
  def tiles = {
    val tiles = palette.images.map(new BasicTile(_))
    val windows = tiles.sliding(subwidth, subwidth).toSeq
    (0 until parts).flatMap(i => windows.drop(i).sliding(1, parts).flatten).flatten
  }
}

object AutoTilePalette {
  def a2(file: File, size: Int = 48) = new AutoFloorTilePalette(file, size)
  def a3(file: File, size: Int = 48) = new AutoWallTilePalette(file, size)
  def a4(file: File, size: Int = 48) = new AutoComboTilePalette(file, size)
  def basic(file: File, size: Int = 48) = new BasicTilePalette(file, size)
  def partitioned(file: File, parts: Int = 2, size: Int = 48) = new PartitionedTilePalette(file, parts, size)
}

import scalafx.Includes._
import scalafx.application.JFXApp
import scalafx.scene._
import scalafx.scene.image.ImageView
import scalafx.scene.layout.TilePane
import scalafx.scene.paint.Color

import Color._

object AutoTileDemo extends JFXApp {
  final val MaxX = 3
  final val MaxY = 3

  val file = raw"C:\Users\dalyj\Documents\Design\Images\AutoTiles\tilea2.png"
  val palette = new AutoFloorTilePalette(new File(file), 32)
  val tile = palette(2, 0)
  val imgs = for (y <- 0 to MaxY; x <- 0 to MaxX) yield {
    val dirs = (x, y) match {
      case (0, 0) => Set(East, South)
      case (MaxX, MaxY) => Set(West, North)
      case (0, MaxY) => Set(East, North)
      case (MaxX, 0) => Set(West, South)
      case (0, _) => Set(East, North, South)
      case (MaxX, _) => Set(West, North, South)
      case (_, 0) => Set(East, West, South)
      case (_, MaxY) => Set(East, West, North)
      case (_, _) => Set(East, West, North, South)
    }
    tile.build(dirs)
  }
  val tiles = imgs.map(i => new ImageView(i.croppedImage()).delegate)
  
  
  val pane = new TilePane(){
        prefColumns = MaxX + 1
        prefRows = MaxY + 1
      }
  pane.children ++= tiles
  pane.children += new ImageView(tile.icon.croppedImage())
  
  stage = new JFXApp.PrimaryStage {
    title.value = "Hello Stage"
    width = 600
    height = 450
    scene = new Scene {
      fill = LightGreen
      content = pane 
    }
  }
  
}