package fmon.stat

import scala.collection.mutable.{Map => MutMap}
import scala.reflect.runtime.universe._
import scala.tools.reflect.ToolBox
import scala.util.Random

import scala.io.Source

import fmon._
import fmon.battle.msg._
import fmon.util._
import EffectType.Volatile

object EffectType extends Enumeration {
  
  val NonVolatile, Volatile, AbilityEffect, MoveEffect, ItemEffect, Field, Weather = Value
  
  def parse(s : String, default : EffectType) = s match {
    case "ability" => AbilityEffect
    case "item" => ItemEffect
    case "move" => MoveEffect
    case "status" | "Status" | "nonvolatile" => NonVolatile
    case "volatile" => Volatile
    case "weather" => Weather
    case _ => default
  }
}

trait Effect {
  def name: String
  def id: String
  def effectType: EffectType
  
  def isItem = effectType == EffectType.ItemEffect
  def isMove = effectType == EffectType.MoveEffect
  def isAbility = effectType == EffectType.AbilityEffect
}

case class EffectSource(mon: Monster, effect: Effect) {
  def name = effect.name
  def id = effect.id
  
  def isItem = effect.isItem
  def isMove = effect.isMove
  def isAbility = effect.isAbility
  
  def triggerMsg = s"[${mon}'s ${effect}]"
}


class Status(template: StatusTemplate) extends Effect {
  def name = template.name
  def id = template.id
  def effectType = template.effectType
  def onStart(mon: Monster, source: EffectSource)(implicit reader: SignalConsumer, rng: Random) = template.onStart(this, mon, source, reader, rng)
  def onEnd(mon: Monster)(implicit reader: SignalConsumer, rng: Random) = template.onEnd(this, mon, reader, rng)
  def onModifyStat(mon: Monster, stat: Stat) = template.onModifyStat(this, mon, stat)
  // val onBeforeMovePriority : Int
  def onBeforeMove(mon: Monster, move: MoveTurn, target: Monster)(implicit reader: SignalConsumer, rng: Random) = template.onBeforeMove(this, mon, move, target, reader, rng)
  // val onModifyMove
  // val onHit
  def onResidualOrder = template.onResidualOrder
  def onResidual(mon: Monster)(implicit reader: SignalConsumer, rng: Random) = template.onResidual(this, mon, reader, rng)
  // val onSwitchIn
  def onLockMove(mon: Monster)(implicit reader: SignalConsumer, rng: Random) = template.onLockMove(this, mon, reader, rng)

  val intData: MutMap[String, Int] = MutMap[String, Int]()
  val strData = MutMap[String, String]()
  
  override def toString = name
}

abstract class StatusTemplate {
  val name: String
  val id: String
  val effectType: EffectType
  val onStart: (Status, Monster, EffectSource, SignalConsumer, Random) => Unit
  val onEnd: (Status, Monster, SignalConsumer, Random) => Unit
  val onModifyStat: (Status, Monster, Stat) => Fraction
  // val onBeforeMovePriority : Int
  val onBeforeMove: (Status, Monster, MoveTurn, Monster, SignalConsumer, Random) => Boolean
  // val onModifyMove
  // val onHit
  val onResidualOrder: Int
  val onResidual: (Status, Monster, SignalConsumer, Random) => Unit
  // val onSwitchIn
  val onLockMove: (Status, Monster, SignalConsumer, Random) => Option[String]

  def build = new Status(this)
  
  override def toString = name
}

case class StatusToken(
  val name:            String,
  val id:              String,
  val effectType:      String,
  val onStart:         String,
  val onEnd:           String,
  val onBeforeMove:    String,
  val onModifyStat:    String,
  val onResidualOrder: Int,
  val onResidual:      String,
  val onLockMove:      String) {
  def instantiate() = {
    val self = this
    new StatusTemplate {
      val name = self.name
      val id = self.id
      val effectType = EffectType.parse(self.effectType, Volatile)
      val onStart = Status.compileOnStart(self.onStart)
      val onEnd = Status.compileOnEnd(self.onEnd)
      val onBeforeMove = Status.compileOnBeforeMove(self.onBeforeMove)
      val onModifyStat = Status.compileOnModifyStat(self.onModifyStat)
      val onResidualOrder = self.onResidualOrder
      val onResidual = Status.compileOnResidual(self.onResidual)
      val onLockMove = Status.compileOnLockMove(self.onLockMove)
    }
  }
}

object Status {
  private var statuses = Map[String, StatusTemplate]()
  val tokens = YamlHelper.extractMap[StatusToken](Source.fromInputStream(Status.getClass.getResourceAsStream("data/statuses.yaml")))

  def apply(name: String): StatusTemplate = {
    if (!statuses.contains(name)) {
      statuses = statuses.updated(name, tokens(name).instantiate())
    }
    statuses(name)
  }

  private val tb = runtimeMirror(getClass.getClassLoader).mkToolBox()
  private val header = """
    import scala.util.Random
    import fmon.battle.msg._
    import fmon._
    import fmon.stat._
    import fmon.stat.MoveType._
    import fmon.stat.Statistic._
    import fmon.util.Fraction
    """
  private val helpers = """
    implicit val dice = rng
    implicit val consumer = reader
    def msg(text: String): Unit = {
      reader ! Message(text)
    }
    """

  def compileOnStart(code: String): (Status, Monster, EffectSource, SignalConsumer, Random) => Unit = {
    if (code == null) {
      (_, _, _, _, _) => ()
    } else {
      val tree = tb.parse(
        s"""
           |$header
           |def onStart(self:Status, mon: Monster, source: EffectSource, reader: SignalConsumer, rng: Random) = {
           |  $helpers
           |  $code
           |}
           |onStart _
          """.stripMargin)
      val f = tb.compile(tree)
      val wrapper = f()
      wrapper.asInstanceOf[(Status, Monster, EffectSource, SignalConsumer, Random) => Unit]
    }
  }
  
  def compileOnEnd(code: String): (Status, Monster, SignalConsumer, Random) => Unit = {
    if (code == null) {
      (_, _, _, _) => ()
    } else {
      val tree = tb.parse(
        s"""
           |$header
           |def onStart(self:Status, mon: Monster, reader: SignalConsumer, rng: Random) = {
           |  $helpers
           |  $code
           |}
           onStart _
          """.stripMargin)
      val f = tb.compile(tree)
      val wrapper = f()
      wrapper.asInstanceOf[(Status, Monster, SignalConsumer, Random) => Unit]
    }
  }

  def compileOnBeforeMove(code: String): (Status, Monster, MoveTurn, Monster, SignalConsumer, Random) => Boolean = {
    if (code == null) {
      (_, _, _, _, _, _) => true
    } else {
      val tree = tb.parse(
        s"""
          $header
          def onBeforeMove(self: Status, mon: Monster, move: MoveTurn, target: Monster, reader: SignalConsumer, rng: Random): Boolean = {
            $helpers
            $code
          }
          onBeforeMove _
          """)
          
      val f = tb.compile(tree)
      val wrapper = f()
      wrapper.asInstanceOf[(Status, Monster, MoveTurn, Monster, SignalConsumer, Random) => Boolean]
    }
  }

  def compileOnModifyStat(code: String): (Status, Monster, Stat) => Fraction = {
    if (code == null) {
      (_, _, _) => 1.frac
    } else {
      val tree = tb.parse(
        s"""
        $header
        def onModifyStat(self: Status, mon: Monster, stat: Stat): Fraction = {
          $code
        }
        onModifyStat _
        """)
      val f = tb.compile(tree)
      val wrapper = f()
      wrapper.asInstanceOf[(Status, Monster, Stat) => Fraction]
    }
  }

  def compileOnResidual(code: String): (Status, Monster, SignalConsumer, Random) => Unit = {
    if (code == null) {
      (_, _, _, _) => ()
    } else {
      val tree = tb.parse(
        s"""
           |$header
           |def onResidual(self: Status, mon: Monster, reader: SignalConsumer, rng: Random) = {
           |  $helpers
           |  $code
           |}
           |onResidual _
          """.stripMargin)
      val f = tb.compile(tree)
      val wrapper = f()

      wrapper.asInstanceOf[(Status, Monster, SignalConsumer, Random) => Unit]
    }
  }
  
  def compileOnLockMove(code: String): (Status, Monster, SignalConsumer, Random) => Option[String] = {
    if (code == null) {
      (_, _, _, _) => None
    } else {
      val tree = tb.parse(
        s"""
           |$header
           |def onLockMove(self: Status, mon: Monster, reader: SignalConsumer, rng: Random) = {
           |  $helpers
           |  val result = {$code}
           |  Some(result)
           |}
           |onLockMove _
          """.stripMargin)
      val f = tb.compile(tree)
      val wrapper = f()

      wrapper.asInstanceOf[(Status, Monster, SignalConsumer, Random) => Option[String]]
    }
  }
}